from typing import Coroutine, Dict, Any, Tuple, Optional, Type, List, Callable
from functools import partial
from time import time
from motor.motor_asyncio import AsyncIOMotorCursor
from bson.objectid import ObjectId
from .base_model import BaseModel, TPydanticModel
from ..db import ObjectsCursor, Shard
from ..errors import ModelDestroyed
from ..context import ctx
from ..types import TModel, KwArgs
from ..util import resolve_id
from ..decorators import save_required


class StorableModel(BaseModel):

    @staticmethod
    def _db() -> Shard:
        return ctx.db.meta

    async def _save_to_db(self) -> None:
        await self._db().save_obj(self)

    async def update(self: TModel,
                     data: Dict[str, Any],
                     *,
                     skip_callback: bool = False,
                     invalidate_cache: bool = True) -> TModel:
        for field, descriptor in self._fields.items():
            if (
                field in data
                and not descriptor.rejected
                and field != "_id"
            ):
                self.__dict__[field] = data[field]
        return await self.save(skip_callback=skip_callback, invalidate_cache=invalidate_cache)

    async def update_from_pydantic(self: TModel,
                                   data: TPydanticModel,
                                   *,
                                   skip_callback: bool = False,
                                   invalidate_cache: bool = True) -> TModel:
        return await self.update(
            data.dict(exclude_unset=True),
            skip_callback=skip_callback,
            invalidate_cache=invalidate_cache
        )

    @save_required
    async def db_update(self,
                        update: Dict[str, Any],
                        when: Optional[Dict[str, Any]] = None,
                        *,
                        reload: bool = True,
                        invalidate_cache: bool = True) -> bool:
        """
        :param update: MongoDB update query
        :param when: filter query. No update will happen if it does not match
        :param reload: Load the new stat into the object (Caution: if you do not do this
                        the next save() will overwrite updated fields)
        :param invalidate_cache: whether to run cache invalidation for the model if the model is changed
        :return: True if the document was updated. Otherwise - False
        """
        new_data = await self._db().find_and_update_obj(self, update, when)
        if invalidate_cache and new_data:
            await self.invalidate()

        if reload and new_data:
            tmp = self.__class__(new_data)
            self._reload_from_model(tmp)

        return bool(new_data)

    async def _delete_from_db(self) -> None:
        await self._db().delete_obj(self)

    async def _refetch_from_db(self: TModel) -> Optional[TModel]:
        return await self.find_one({"_id": self.id})

    async def reload(self) -> None:
        if self.is_new:
            return
        tmp = await self._refetch_from_db()
        if tmp is None:
            raise ModelDestroyed("model has been deleted from db")
        self._reload_from_model(tmp)

    @classmethod
    def _preprocess_query(cls, query: Dict[str, Any]) -> Dict[str, Any]:
        return query

    @classmethod
    def find(cls: Type[TModel],
             query: Optional[Dict[str, Any]] = None,
             **kwargs: KwArgs) -> ObjectsCursor[TModel]:
        if not query:
            query = {}
        return cls._db().get_objs(
            cls._ctor, cls.collection, cls._preprocess_query(query), **kwargs
        )

    @classmethod
    def aggregate(cls,
                  pipeline: List[Dict[str, Any]],
                  query: Optional[Dict[str, Any]] = None,
                  **kwargs: KwArgs) -> AsyncIOMotorCursor:
        if not query:
            query = {}
        pipeline = [{"$match": cls._preprocess_query(query)}] + pipeline
        return cls._db().get_aggregated(cls.collection, pipeline, **kwargs)

    @classmethod
    def find_projected(cls,
                       query: Optional[Dict[str, Any]] = None,
                       projection: Tuple[str] = ("_id",),
                       **kwargs: KwArgs) -> AsyncIOMotorCursor:
        if not query:
            query = {}
        return cls._db().get_objs_projected(
            cls.collection,
            cls._preprocess_query(query),
            projection=projection,
            **kwargs,
        )
    
    @classmethod
    async def find_ids(cls,
                       query: Optional[Dict[str, Any]] = None,
                       max_count: Optional[int] = None,
                       **kwargs: KwArgs) -> List[ObjectId]:
        objs = await cls.find_projected(query, **kwargs).to_list(max_count)
        return [obj["_id"] for obj in objs]

    @classmethod
    async def find_one(cls: Type[TModel], query: Dict[str, Any], **kwargs: KwArgs) -> Optional[TModel]:
        return await cls._db().get_obj(
            cls._ctor, cls.collection, cls._preprocess_query(query), **kwargs
        )

    @classmethod
    async def get(cls: Type[TModel],
                  expression: str | ObjectId | None,
                  raise_if_none: Optional[Exception] = None) -> Optional[TModel]:
        if expression is None:
            if raise_if_none:
                raise raise_if_none
            return None

        rexp = resolve_id(expression)
        if isinstance(rexp, ObjectId):
            query = {"_id": rexp}
        else:
            query = {cls.KEY_FIELD: str(rexp)}
        res = await cls.find_one(query)
        if res is None and raise_if_none is not None:
            raise raise_if_none
        return res

    @classmethod
    async def count(cls, query: Optional[Dict[str, Any]] = None) -> int:
        if query is None:
            query = {}
        return await cls._db().count_docs(cls.collection, query)

    @classmethod
    async def destroy_all(cls) -> None:
        await cls._db().delete_query(cls.collection, cls._preprocess_query({}))

    @classmethod
    async def destroy_many(cls, query: Dict[str, Any]) -> None:
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        await cls._db().delete_query(cls.collection, cls._preprocess_query(query))

    @classmethod
    async def update_many(cls, query: Dict[str, Any], attrs: Dict[str, Any]) -> None:
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        await cls._db().update_query(cls.collection, cls._preprocess_query(query), attrs)

    @classmethod
    async def cache_get(cls: Type[TModel],
                        expression: str | None,
                        raise_if_none: Optional[Exception] = None) -> Optional[TModel]:
        if expression is None:
            if raise_if_none:
                raise raise_if_none
            return None
        cache_key = f"{cls.collection}.{expression}"
        getter = partial(cls.get, expression, raise_if_none)
        return await cls._cache_get(cache_key, getter)

    @classmethod
    async def _cache_get(cls: Type[TModel],
                         cache_key: str,
                         getter: partial[Coroutine[None, None, TModel | None]],
                         ctor: Optional[Callable[..., TModel]] = None) -> Optional[TModel]:
        t1 = time()
        if not ctor:
            ctor = cls

        if await ctx.cache_l1.has(cache_key):
            data = await ctx.cache_l1.get(cache_key)
            td = time() - t1
            ctx.log.debug(
                "%s L1 hit %s %.3f secs", ctx.cache_l1.__class__.__name__, cache_key, td
            )
            return ctor(data)

        if await ctx.cache_l2.has(cache_key):
            data = await ctx.cache_l2.get(cache_key)
            await ctx.cache_l1.set(cache_key, data)
            td = time() - t1
            ctx.log.debug(
                "%s L2 hit %s %.3f secs", ctx.cache_l2.NAME, cache_key, td
            )
            return ctor(data)

        obj: Optional[TModel] = await getter()
        if obj:
            data = obj.to_dict(include_restricted=True, convert_id=True)
            await ctx.cache_l2.set(cache_key, data)
            await ctx.cache_l1.set(cache_key, data)

        td = time() - t1
        ctx.log.debug("%s miss %s %.3f secs", ctx.cache_l2.NAME, cache_key, td)
        return obj

    async def invalidate(self, **kwargs: KwArgs) -> None:
        for field in self._cache_key_fields:
            value = self._initial_state.get(field)
            if value:
                cache_key = f"{self.collection}.{value}"
                await self._invalidate(cache_key)
        if not self.is_new:
            for method in self._cached_methods:
                cache_key = f"{self.collection}.{self.id}.{method}"
                await self._invalidate(cache_key)

    @staticmethod
    async def _invalidate(cache_key: str) -> None:
        if await ctx.cache_l1.delete(cache_key):
            ctx.log.debug("%s delete %s", ctx.cache_l1.NAME, cache_key)
        if await ctx.cache_l2.delete(cache_key):
            ctx.log.debug("%s delete %s", ctx.cache_l2.NAME, cache_key)



from fastapi import APIRouter, Depends, Body
from croydon import ctx
from croydon.errors import AuthenticationError, BadRequest
from app.extractors import authenticated_user, current_session
from app.models import User, Session
from .response_types.account import AccountMeResponse, LogoutResponse, AuthenticationRequest

account_ctrl = APIRouter(prefix="/api/v1/account")


@account_ctrl.get("/me")
async def me(user: User = Depends(authenticated_user())) -> AccountMeResponse:
    return AccountMeResponse(**user.to_dict())


@account_ctrl.post("/authenticate")
async def authenticate(
        auth_request: AuthenticationRequest = Body(),
        session: Session = Depends(current_session)) -> AccountMeResponse:

    user = await session.user()
    if user:
        raise BadRequest("already authenticated")

    user = await User.get(auth_request.username)
    if user is None or not user.check_password(auth_request.password):
        raise AuthenticationError()
    session.user_id = user.id

    return AccountMeResponse(**user.to_dict())


@account_ctrl.post("/logout")
async def logout(session: Session = Depends(current_session)) -> LogoutResponse:
    session.user_id = None
    return LogoutResponse(detail="logged out")



from typing import Optional
from croydon.models.storable_model import StorableModel
from croydon.models.fields import StringField, ReferenceField, DatetimeField
from croydon.util import uuid4_string, now
from .user import User


class Session(StorableModel):

    COLLECTION = "sessions"
    KEY_FIELD = "key"

    key = StringField(default=uuid4_string, required=True, rejected=True, restricted=True)
    user_id: ReferenceField[User] = ReferenceField(reference_model=User, index=True)
    created_at = DatetimeField(default=now, required=True)

    async def user(self) -> Optional["User"]:
        if self.user_id is None:
            return None
        await User.cache_get(self.user_id)



import inspect
from typing import (
    Generic, Callable, Optional, List, Type,
    AsyncIterator, Any, Dict, Sequence, Union, Mapping, Tuple
)
from time import sleep
from datetime import datetime

import pymongo
from bson.objectid import ObjectId, InvalidId
from motor.motor_asyncio import AsyncIOMotorClient
from motor.motor_asyncio import AsyncIOMotorCursor
from motor.motor_asyncio import AsyncIOMotorDatabase
from mongomock_motor import (
    AsyncMongoMockClient, AsyncMongoMockDatabase, AsyncCursor
)
from pymongo import ReturnDocument
from pymongo.results import DeleteResult, UpdateResult
from pymongo.errors import ServerSelectionTimeoutError
from .config import DatabaseConfig
from .cache import TCache
from .types import RT, TModel, KwArgs, QueryExpression, QueryType, CommonDict

MONGO_RETRY_MAX: int = 6
MONGO_RETRY_SLEEP: int = 3


def intercept_pymongo_errors(func: Callable[..., RT]) -> Callable[..., RT]:
    def wrapper(self: "Shard", *args: List[Any], **kwargs: KwArgs) -> RT:
        retries = MONGO_RETRY_MAX
        while True:
            retries -= 1
            try:
                return func(self, *args, **kwargs)
            except ServerSelectionTimeoutError:
                from .context import ctx
                ctx.log.error(
                    "ServerSelectionTimeoutError in db module, retries left = %d",
                    retries,
                )
                if retries == MONGO_RETRY_MAX // 2:
                    ctx.log.error("trying to reinstantiate connection")
                    db_inst = self
                    db_inst.reset_conn()
                elif retries == 0:
                    ctx.log.error(
                        "ServerSelectionTimeoutError max retries exceeded, giving up"
                    )
                    raise
                sleep(MONGO_RETRY_SLEEP)

    return wrapper


class ObjectsCursor(Generic[TModel], AsyncIterator[TModel]):
    cursor: Union[AsyncIOMotorCursor, AsyncCursor]

    def __init__(self, cursor: Union[AsyncIOMotorCursor, AsyncCursor], ctor: Type[TModel], query: QueryType,
                 shard_id: Optional[str] = None) -> None:
        self.ctor = ctor
        self.cursor = cursor
        self.stored_query = query
        self._shard_id = shard_id

    async def all(self, max_length: Optional[int] = None) -> List[TModel]:
        return [self.ctor(x) for x in await self.cursor.to_list(max_length)]

    async def all_dict(self,
                       max_length: Optional[int] = None,
                       fields: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        return [self.ctor(x).to_dict(fields=fields) for x in await self.cursor.to_list(max_length)]

    def limit(self, limit: int) -> "ObjectsCursor[TModel]":
        self.cursor.limit(limit)
        return self

    def skip(self, skip: int) -> "ObjectsCursor[TModel]":
        self.cursor.skip(skip)
        return self

    def sort(self, key_or_list: Union[str, Sequence[Tuple[str, Union[int, str, Mapping[str, Any]]]]],
             direction: Optional[Union[int, str]] = None) -> "ObjectsCursor[TModel]":
        if isinstance(key_or_list, str) and direction is None:
            direction = pymongo.ASCENDING
        self.cursor.sort(key_or_list, direction)
        return self

    async def count(self) -> int:
        # usually self.ctor contains a classmethod `ctor` of a StorableModel
        # hence this hack
        ctor = self.ctor.__self__
        return await ctor.count(self.stored_query)

    def __aiter__(self) -> AsyncIterator["TModel"]:
        return self

    async def __anext__(self) -> "TModel":
        item = await self.cursor.__anext__()
        if self._shard_id:
            return self.ctor(shard_id=self._shard_id, attrs=item)
        else:
            return self.ctor(attrs=item)

    def __getattr__(self, item: str) -> Any:
        return getattr(self.cursor, item)


class Shard:
    _config: DatabaseConfig
    _client: Optional[Union[AsyncIOMotorClient, AsyncMongoMockClient]]
    _conn: Optional[Union[AsyncIOMotorDatabase, AsyncMongoMockDatabase]]
    _mock: bool

    def __init__(self, config: DatabaseConfig, mock: bool = False) -> None:
        self._config = config
        self._client = None
        self._conn = None
        self._mock = mock

    @property
    def _dbname(self) -> str:
        return self._config.dbname

    def get_client(self) -> Union[AsyncIOMotorClient, AsyncMongoMockClient]:
        if not self._client:
            if self._mock:
                self._client = AsyncMongoMockClient(self._config.uri)
            else:
                self._client = AsyncIOMotorClient(self._config.uri)
        return self._client

    def reset_conn(self) -> None:
        self._client = None
        self._conn = None

    def init_conn(self) -> None:
        from .context import ctx
        if not self._mock:
            ctx.log.info("creating a new mongo connection")
        client = self.get_client()
        self._conn = client[self._dbname]

    @property
    def conn(self) -> Union[AsyncIOMotorDatabase, AsyncMongoMockDatabase]:
        if self._conn is None:
            self.init_conn()
        return self._conn

    @intercept_pymongo_errors
    async def get_obj(self,
                      ctor: Callable[..., "TModel"],
                      collection: str,
                      query: QueryExpression) -> Optional["TModel"]:
        if not isinstance(query, dict):
            try:
                query = {"_id": ObjectId(query)}
            except InvalidId:
                pass
        data = await self.conn[collection].find_one(query)
        if data:
            return ctor(attrs=data)

        return None

    @intercept_pymongo_errors
    async def get_obj_id(self, collection: str, query: QueryType) -> Optional[ObjectId]:
        obj = await self.conn[collection].find_one(
            query, projection=()
        )
        if obj:
            return obj["_id"]
        return None

    @intercept_pymongo_errors
    def get_objs(self,
                 ctor: Callable[..., "TModel"],
                 collection: str,
                 query: QueryType,
                 **kwargs: Dict[str, Any]) -> ObjectsCursor[TModel]:
        cursor = self.conn[collection].find(query, **kwargs)
        return ObjectsCursor(cursor, ctor, query)

    @intercept_pymongo_errors
    def get_objs_projected(self,
                           collection: str,
                           query: CommonDict,
                           projection: CommonDict, **kwargs: KwArgs) -> AsyncIOMotorCursor:
        cursor = self.conn[collection].find(query, projection=projection, **kwargs)
        return cursor

    @intercept_pymongo_errors
    def get_aggregated(self, collection: str, pipeline: List[CommonDict], **kwargs: KwArgs) -> AsyncIOMotorCursor:
        cursor = self.conn[collection].aggregate(pipeline, **kwargs)
        return cursor

    @intercept_pymongo_errors
    async def count_docs(self, collection: str, query: CommonDict, **kwargs: KwArgs) -> int:
        coll = self.conn[collection]
        if query == {}:
            return await coll.estimated_document_count(**kwargs)
        else:
            return await coll.count_documents(query, **kwargs)

    def get_objs_by_field_in(self,
                             ctor: Type[TModel],
                             collection: str,
                             field: str,
                             values: List[str],
                             **kwargs: KwArgs) -> ObjectsCursor[TModel]:
        return self.get_objs(ctor, collection, {field: {"$in": values}}, **kwargs)

    @intercept_pymongo_errors
    async def save_obj(self, obj: "TModel") -> None:
        if obj.is_new:
            # object to_dict() method should always return all fields
            data = obj.to_dict(include_restricted=True, convert_id=True)
            # although with the new object we shouldn"t pass _id=null to mongo
            del data["_id"]
            insert_result = await self.conn[obj.collection].insert_one(data)
            obj.id = insert_result.inserted_id
        else:
            await self.conn[obj.collection].replace_one(
                {"_id": obj.id},
                obj.to_dict(include_restricted=True, convert_id=True),
                upsert=True,
            )

    @intercept_pymongo_errors
    async def delete_obj(self, obj: "TModel") -> None:
        if obj.is_new:
            return
        await self.conn[obj.collection].delete_one({"_id": obj.id})

    @intercept_pymongo_errors
    async def find_and_update_obj(self, obj: "TModel", update: CommonDict, when: Optional[CommonDict] = None) -> None:
        query = {"_id": obj.id}
        if when:
            assert "_id" not in when
            query.update(when)

        new_data = await self.conn[obj.collection].find_one_and_update(
            query, update, return_document=ReturnDocument.AFTER
        )
        return new_data

    @intercept_pymongo_errors
    async def delete_query(self, collection: str, query: CommonDict) -> DeleteResult:
        return await self.conn[collection].delete_many(query)

    @intercept_pymongo_errors
    async def update_query(self, collection: str, query: CommonDict, update: CommonDict) -> UpdateResult:
        return await self.conn[collection].update_many(query, update)

    # SESSIONS
    @intercept_pymongo_errors
    async def get_session(self, sid: str, collection: str = "sessions") -> CommonDict:
        return await self.conn[collection].find_one({"sid": sid})

    @intercept_pymongo_errors
    async def update_session(self, sid: str, data: CommonDict, expiration: datetime,
                             collection: str = "sessions") -> None:
        await self.conn[collection].update_one(
            {"sid": sid}, {"$set": {"sid": sid, "data": data, "expiration": expiration}}, True
        )

    @intercept_pymongo_errors
    async def cleanup_sessions(self, collection: str = "sessions") -> int:
        result = await self.conn[collection].delete_one({"expiration": {"$lt": datetime.now()}})
        return result.deleted_count


class DB:
    MONGODB_INFO_FIELDS = (
        "allocator",
        "bits",
        "debug",
        "gitVersion",
        "javascriptEngine",
        "maxBsonObjectSize",
        "modules",
        "ok",
        "openssl",
        "storageEngines",
        "sysInfo",
        "version",
        "versionArray",
    )

    __meta: Optional[Shard]

    def __init__(self,
                 *,
                 config: DatabaseConfig,
                 mock: bool = False) -> None:
        self.reconfigure(config, mock)

    def reconfigure(self, config: DatabaseConfig, mock: bool = False) -> None:
        self.__meta = Shard(config, mock=mock)

    @property
    def meta(self) -> Shard:
        return self.__meta



import inspect
from typing import Callable, Awaitable, Any, List
from functools import wraps, update_wrapper
from time import time

from .types import KwArgs, TModel, TModelType, RT
from .errors import ObjectSaveRequired


def save_required(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    save_required decorator is meant to decorate model methods that
    require a model instance to be saved first.

    For example, StorableModel's own db_update method makes no sense
    until the instance is persisted in the database
    """
    @wraps(func)
    def wrapper(self, *args: List[Any], **kwargs: KwArgs) -> Any:
        if self.is_new:
            raise ObjectSaveRequired("This object must be saved first")
        return func(self, *args, **kwargs)

    return wrapper


class api_field:
    """
    api_field is a decorator for model no-args methods (including async),
    allowing API users to access certain methods, i.e. implementing computed fields
    """

    def __init__(self, fn):
        if not inspect.isfunction(fn):
            raise RuntimeError("only functions and methods can be decorated with api_field")
        self.fn = fn

    def __set_name__(self, owner, name):
        from .models.fields import ComputedField
        fd = ComputedField(is_async=inspect.iscoroutinefunction(self.fn))

        # computed_fields are inherited from MetaModel down the class tree
        # here we make sure every descendant does not modify parent's computed fields
        owner._computed_fields = owner._computed_fields.copy()
        owner._computed_fields[name] = fd
        setattr(owner, name, self.fn)


class model_cached_method:
    """
    model_cached_method decorates model methods. The wrapper provided
    generates a unique cache key out of
    - model class name
    - model id
    - method name

    e.g. User.6433d2bfd5a353b0a6822d7f.full_name

    and tries to fetch the value from cache. If value is not there, the
    original method is called and the value is stored to cache before
    being returned

    The methods wrapped with this decorator are automatically added to a
    class property called _cached_methods: set. This allows StorableModel.invalidate
    method to invalidate these cached values along with other model-provided cache.
    """

    orig_func: Callable[[TModel, ...], Awaitable[RT]] = None

    def __init__(self, func: Callable[[TModel, ...], Awaitable[RT]]) -> None:
        self.orig_func = func

    def __set_name__(self, owner: TModelType, name: str) -> None:

        @wraps(self.orig_func)
        async def wrapper(this: TModel, *args, **kwargs) -> RT:
            from .context import ctx
            cache_key = f"{this.collection}.{this.id}.{self.orig_func.__name__}"

            t1 = time()
            value = await ctx.cache_l1.get(cache_key)
            if value:
                td = time() - t1
                ctx.log.debug(
                    "%s L1 hit %s %.3f secs", ctx.cache_l1.NAME, cache_key, td
                )
                return value

            t1 = time()
            value = await ctx.cache_l2.get(cache_key)
            if value:
                td = time() - t1
                ctx.log.debug(
                    "%s L2 hit %s %.3f secs", ctx.cache_l2.NAME, cache_key, td
                )
                await ctx.cache_l1.set(cache_key, value)
                return value

            t1 = time()
            value = await self.orig_func(this, *args, **kwargs)
            td = time() - t1
            ctx.log.debug("%s miss %s %.3f secs", ctx.cache_l2.NAME, cache_key, td)
            await ctx.cache_l2.set(cache_key, value)
            await ctx.cache_l1.set(cache_key, value)

            return value

        owner._cached_methods.add(name)
        setattr(owner, name, wrapper)



from ..models.storable_model import StorableModel
from ..models.fields import StringField, Field, ObjectIdField
from ..db import ObjectsCursor
from ..decorators import api_field, model_cached_method
from .mongo_mock_test import MongoMockTest
from croydon.cache import TraceCache


CALLABLE_DEFAULT_VALUE = 4


def callable_default():
    return CALLABLE_DEFAULT_VALUE


class TestModel(StorableModel):
    field1 = StringField(rejected=True, index=True, default="default_value")
    field2 = Field(required=True)
    field3 = Field(required=True, default="required_default_value")
    callable_default_field = Field(default=callable_default)


class TestStorableModel(MongoMockTest):

    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        await TestModel.destroy_all()

    async def asyncTearDown(self) -> None:
        await TestModel.destroy_all()
        await super().asyncTearDown()

    async def test_defaults(self):
        model = TestModel()
        self.assertEqual("default_value", model.field1)
        self.assertEqual("required_default_value", model.field3)
        self.assertEqual(CALLABLE_DEFAULT_VALUE, model.callable_default_field)

    async def test_eq(self):
        model = TestModel({"field2": "mymodel"})
        await model.save()
        model2 = await TestModel.find_one({"field2": "mymodel"})
        self.assertEqual(model, model2)

    async def test_reject_on_update(self):
        model = TestModel.create(field1="original_value", field2="mymodel_reject_test")
        await model.save()
        id_ = model.id
        await model.update({"field1": "new_value"})
        model = await TestModel.find_one({"_id": id_})
        self.assertEqual(model.field1, "original_value")

    async def test_update_many(self):
        model1 = TestModel.create(field1="original_value", field2="mymodel_update_test")
        await model1.save()
        model2 = TestModel.create(field1="original_value", field2="mymodel_update_test")
        await model2.save()
        model3 = TestModel.create(field1="do_not_modify", field2="mymodel_update_test")
        await model3.save()

        await TestModel.update_many(
            {"field1": "original_value"}, {"$set": {"field2": "mymodel_updated"}}
        )
        await model1.reload()
        await model2.reload()
        await model3.reload()

        self.assertEqual(model1.field2, "mymodel_updated")
        self.assertEqual(model2.field2, "mymodel_updated")
        self.assertEqual(model3.field2, "mymodel_update_test")

    async def test_invalidate(self):
        from croydon import ctx
        tc: TraceCache = ctx.cache_l1

        class Model(StorableModel):
            field1 = StringField()
            KEY_FIELD = "field1"

        model = Model({"field1": "value"})
        await model.save()

        tc.called_once("delete", "model.value")
        tc.reset()

        await model.save()
        tc.called_once("delete", f"model.{model.id}")
        tc.called_once("delete", f"model.value")

    async def test_invalidate_cached_method(self):
        from croydon import ctx
        tc: TraceCache = ctx.cache_l1

        class User(StorableModel):
            first_name = StringField()
            last_name = StringField()

            KEY_FIELD = "last_name"

            @model_cached_method
            async def full_name(self):
                return f"{self.first_name} {self.last_name}"

        user = User.create(first_name="Bob", last_name="Dilan")
        await user.save()

        tc.called_once("delete", "user.Dilan")

        _ = await user.full_name()
        tc.called_once("get", f"user.{user.id}.full_name")
        tc.reset()

        await user.save()
        tc.called_once("delete", "user.Dilan")
        tc.called_once("delete", f"user.{user.id}")
        tc.called_once("delete", f"user.{user.id}.full_name")

    async def test_update(self):
        model = TestModel.create(field1="original_value", field2="mymodel_update_test")
        await model.save()
        id_ = model.id
        await model.update({"field2": "mymodel_updated"})
        model = await TestModel.find_one({"_id": id_})
        self.assertEqual(model.field2, "mymodel_updated")

    async def test_count(self):
        class Model(StorableModel):
            a = Field()

        for i in range(100):
            m = Model.create(a=i)
            await m.save()

        cur = Model.find({})
        count = await cur.count()
        self.assertEqual(count, 100)

        cur = Model.find({"a": {"$lt": 10}})
        count = await cur.count()
        self.assertEqual(count, 10)

    async def test_to_dict_ext_cursor_properties(self):
        class DepModel(StorableModel):
            a = Field()
            master_id = ObjectIdField(required=True)

        class MasterModel(StorableModel):
            name = Field()

            @api_field
            def deps(self) -> ObjectsCursor["DepModel"]:
                return DepModel.find({"master_id": self.id})

        master = MasterModel.create(name="master")
        await master.save()

        for i in range(10):
            dep = DepModel.create(a=i, master_id=master.id)
            await dep.save()

        # check normal method usage
        deps = await master.deps().all()
        self.assertEqual(len(deps), 10)

        # check deps are accessible via API
        dct = await master.to_dict_ext(fields=["id", "name", "deps"])
        self.assertEqual(len(dct["deps"]), 10)

    async def test_find_by_none_raises(self):
        with self.assertRaises(ValueError):
            _ = await TestModel.get(None, ValueError("value error"))
        with self.assertRaises(ValueError):
            _ = await TestModel.cache_get(None, ValueError("value error"))

    async def test_find_ids(self):
        tma = TestModel.create(field2="a")
        await tma.save()
        tmb = TestModel.create(field2="b")
        await tmb.save()
        tmc = TestModel.create(field2="c")
        await tmc.save()

        self.assertCountEqual([tma.id, tmb.id, tmc.id], await TestModel.find_ids({}))
        self.assertCountEqual([tma.id], await TestModel.find_ids({"field2": "a"}))



import os
import IPython
from croydon import ctx
from croydon.command import Command


class Shell(Command):

    NAME = "shell"
    HELP = "run project-aware python shell"
    ASYNC_RUN = False

    def run_sync(self):
        from app import app
        app.initialise()

        rcfile = ".shellrc.py"
        if not os.path.isfile(rcfile):
            rcfile = os.path.join(ctx.project_dir, ".shellrc.py")

        if os.path.isfile(rcfile):
            with open(rcfile) as rcf:
                try:
                    exec(rcf.read())
                except Exception:
                    from traceback import print_exc

                    print("Error running .shellrc.py script!")
                    print_exc()

        IPython.embed(using='asyncio', colors="neutral")



from unittest import IsolatedAsyncioTestCase
from typing import Type, TypeVar, Callable, Awaitable, Any
from datetime import datetime
from ..models.fields import (
    StringField,
    IntField,
    FloatField,
    ListField,
    DictField,
    BoolField,
    DatetimeField,
)
from ..models.base_model import BaseModel
from ..errors import ValidationError

TException = TypeVar("TException", bound=Exception)
TExceptionType = Type[TException]


class TestValidators(IsolatedAsyncioTestCase):

    async def test_required(self):
        class Model(BaseModel):
            field = StringField(required=True)

        model = Model()
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = "string"
        await model.validate_all()

    async def test_string_field(self):
        class Model(BaseModel):
            field = StringField()

        model = Model({"field": 1})
        with self.assertRaises(ValidationError):
            await model.validate_all()

        model.field = "string"
        await model.validate_all()

        model.field = "   auto-trimmed   "
        self.assertEqual(model.field, "auto-trimmed")

        class Model(BaseModel):
            field = StringField(auto_trim=False)

        model = Model({"field": "    no-auto-trim    "})
        self.assertEqual(model.field, "    no-auto-trim    ")

        class Model(BaseModel):
            field = StringField(min_length=3, max_length=5, re_match=r"^\d+$")

        model = Model({"field": "1"})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = "123456"
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = "1234a"
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = "1234"
        await model.validate_all()

    async def test_list_field(self):
        class Model(BaseModel):
            field = ListField()

        model = Model({"field": 1})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = [1, 2, 3]
        await model.validate_all()

        class Model(BaseModel):
            field = ListField(min_length=3, max_length=5)

        model = Model({"field": [1]})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = [1, 2, 3, 4, 5, 6]
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = [1, 2, 3, 4]
        await model.validate_all()

    async def test_bool_field(self):
        class Model(BaseModel):
            field = BoolField()

        model = Model({"field": 1})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = False
        await model.validate_all()

    async def test_int_field(self):
        class Model(BaseModel):
            field = IntField()

        model = Model({"field": "string"})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = -34
        await model.validate_all()

        class Model(BaseModel):
            field = IntField(min_value=0, max_value=10)

        model = Model({"field": -34})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = 34
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = 5
        await model.validate_all()

    async def test_float_field(self):
        class Model(BaseModel):
            field = FloatField()

        model = Model({"field": "string"})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = -34
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = 28.0
        await model.validate_all()

        class Model(BaseModel):
            field = FloatField(min_value=0, max_value=10)

        model = Model({"field": -15.5})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = 15.5
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = 5.3
        await model.validate_all()

    async def test_dict_field(self):
        class Model(BaseModel):
            field = DictField()

        model = Model({"field": 1})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = {}
        await model.validate_all()

    async def test_datetime_field(self):
        class Model(BaseModel):
            field = DatetimeField()

        model = Model({"field": 1})
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = datetime.now()
        await model.validate_all()

    async def test_choices(self):
        class Model(BaseModel):
            field = StringField(required=True, choices=["a", "b", "c"])

        class ModelNotRequired(BaseModel):
            field = StringField(required=False, choices=["a", "b", "c"])

        model = Model({"field": "a"})
        await model.validate_all()
        model.field = "b"
        await model.validate_all()
        model.field = "c"
        await model.validate_all()
        model.field = "d"
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = None
        with self.assertRaises(ValidationError):
            await model.validate_all()

        model = ModelNotRequired({"field": "a"})
        await model.validate_all()
        model.field = "b"
        await model.validate_all()
        model.field = "c"
        await model.validate_all()
        model.field = "d"
        with self.assertRaises(ValidationError):
            await model.validate_all()
        model.field = None
        await model.validate_all()



from typing import Optional
from croydon.models import StorableModel
from croydon.models.fields import StringField, DatetimeField, ReferenceField
from croydon.models.index import Index, IndexDirection, IndexKey
from croydon.util import now, uuid4_string
from .user import User


class Token(StorableModel):

    COLLECTION = "tokens"
    KEY_FIELD = "token"
    INDEXES = [
        Index(
            keys=[
                IndexKey("user_id", IndexDirection.ASCENDING),
                IndexKey("token_type", IndexDirection.ASCENDING),
            ]
        )
    ]

    token_type = StringField(default="auth", required=True, rejected=True)
    token = StringField(default=uuid4_string, required=True, rejected=True, unique=True)
    created_at = DatetimeField(default=now, required=True, rejected=True)
    updated_at = DatetimeField(default=now, required=True, rejected=True)
    description = StringField(default="")
    user_id: ReferenceField[User] = ReferenceField(
        reference_model=User,
        required=True,
        rejected=True,
        restricted=True
    )

    def touch(self):
        self.updated_at = now()

    async def user(self) -> Optional[User]:
        return await User.get(self.user_id)



import json
from typing import Dict, Optional, Generic, Self
from datetime import datetime

from ..util import uuid4_string, now
from ..context import ctx
from .types import TTaskData, TBaseTaskType, Message, TState


class BaseTask(Generic[TTaskData]):

    TYPE: str = "base"
    REGISTRY: Dict[str, TBaseTaskType] = {}

    id: str
    data: TTaskData
    created_at: datetime
    state: TState
    status_message: str
    _received_by: Optional[str]

    def __init__(self,
                 data: TTaskData,
                 task_id: Optional[str] = None,
                 created_at: Optional[datetime] = None,
                 state: TState = "new",
                 status_message: str = "") -> None:
        self.id = task_id or uuid4_string()
        self.data = data
        self.created_at = created_at or now()
        self.state = state
        self.status_message = status_message

    def to_message(self) -> Message:
        return Message(
            id=self.id,
            type=self.TYPE,
            data=json.dumps(self.data),
            created_at=self.created_at,
            state=self.state,
            status_message=self.status_message,
        )

    def _get_received_by(self) -> Optional[str]:
        return self._received_by

    def _set_received_by(self, value: str):
        self._received_by = value

    received_by = property(_get_received_by, _set_received_by)

    @property
    def is_received(self) -> bool:
        return self._received_by is not None

    async def publish(self) -> bool:
        return await ctx.queue.enqueue(self)

    async def done(self, status_message: str = "ok") -> None:
        return await ctx.queue.set_task_done(self, status_message)

    @classmethod
    def from_message(cls, msg: Message) -> Self:
        if "data" not in msg:
            ctx.log.error("malformed message, no data field: %s", msg)
        task_id = msg["id"]
        task_type = msg["type"]
        created_at = msg["created_at"]
        state = msg["state"]
        status_message = msg["status_message"]
        data = json.loads(msg["data"])
        if task_type in cls.REGISTRY:
            task_class = cls.REGISTRY[task_type]
        else:
            task_class = cls
        return task_class(
            data=data,
            task_id=task_id,
            created_at=created_at,
            state=state,
            status_message=status_message
        )

    @classmethod
    def register(cls):
        ctx.log.info("Registering task type %s", cls.TYPE)
        BaseTask.REGISTRY[cls.TYPE] = cls

    def __str__(self):
        return (
            f"<{self.__class__.__name__} {self.TYPE} id={self.id} data={json.dumps(self.data)} "
            + f"created_at={self.created_at} state={self.state} status_message=\"{self.status_message}\">"
        )

    def __repr__(self):
        return self.__str__()



from unittest import IsolatedAsyncioTestCase
from ..context import ctx
from ..config import DatabaseConfig
from ..cache import TraceCache, SimpleCache


class MongoMockTest(IsolatedAsyncioTestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        if not ctx.initialised:
            from ..baseapp import BaseApp
            app = BaseApp(".")
            app.initialise()
        ctx.db.reconfigure(DatabaseConfig(), mock=True)

        # aiomcache conflicts with async tests as it seems to run
        # in its own async loop. This leads to "got Future attached to a different loop"
        # errors.
        #
        # Since mongo tests do not use memcached explicitly we override
        # all caches to avoid using MemcachedCache
        ctx._cache_l1 = TraceCache()
        ctx._cache_l2 = SimpleCache()
